
dofile("Base.rte/Constants.lua")
require("Actors/AI/NativeCrabAI")

function Create(self)
	self.AI = NativeTurretAI:Create(self)
	self.Ctrl = self:GetController()
	
	-- Find the gun
	local thisID = self.ID
	for i = 1, MovableMan:GetMOIDCount()-1 do
		if MovableMan:GetRootMOID(i) == thisID then	-- an obj without a parent has ID == RootID
			self.Gun = MovableMan:GetMOFromID(i)
			if self.Gun.ClassName == "HDFirearm" then
				self.Gun = ToAttachable(self.Gun)
				break
			end
		end
	end
	
	-- debug
	self.startPos = self.Pos
end

function Update(self)
	self:GetController():SetState(Controller.MOVE_RIGHT, false)
	self:GetController():SetState(Controller.MOVE_LEFT, false)
	self.Pos = self.startPos	-- Stay!

	if self.Gun and self.Gun:IsAttached() then
		-- Restrict aiming
		if self.HFlipped then
			if self.Gun.RotAngle > 0.22 then
				self.Gun.RotAngle = 0.22
			end
		elseif self.Gun.RotAngle < -0.22 then
			self.Gun.RotAngle = -0.22
		end
		
		-- Animate barrels
		if self.Gun.Sharpness > 0 then
			if self.Gun.Sharpness < 40 then
				if self.Gun.Sharpness % 2 == 0 then
					if self.Gun.Sharpness < 25 then
						self.Gun.Frame = 4
					else
						self.Gun.Frame = 5
					end
				else
					if self.Gun.Sharpness < 25 then
						self.Gun.Frame = 2
					else
						self.Gun.Frame = 3
					end
				end
				
				self.Gun.Sharpness = self.Gun.Sharpness + 2
			else
				self.Gun.Sharpness = 0
			end
		end
	elseif PosRand() < 0.1 then
		self.Health = self.Health - 3
	end
end

function UpdateAI(self)
	self.AI:Update(self)
end

--[[

NailerTurretAI = {}

-- open fire on the selected target
function NailerTurretAI.ShootTarget(self)
	if not MovableMan:IsActor(self.Target) then
		return true
	end
	
	local TargetAvgVel = self.Target.Vel
	local ShootTimer = Timer()
	local aimTime = RangeRand(600, 1000)
	local aimError = RangeRand(-0.2, 0.2)
	local AimPoint
	
	while self.Gun:IsAttached() do
		if self.FirearmIsReady then
			TargetAvgVel = TargetAvgVel * 0.4 + self.Target.Vel * 0.6	-- filter the target's velocity
			
			if not self.flying then
				self.deviceState = ACrab.AIMING
			end
			
			if SceneMan:ShortestDistance(self.Pos, AimPoint or self.Target.Pos, false).Magnitude < 40 then
				AimPoint = self.Target.Pos + self.TargetOffset * 0.9	-- move the aimpoint closer to the center of the target at close ranges
			else
				AimPoint = self.Target.Pos + self.TargetOffset
			end
			
			aim = HumanAI.GetHitAngle(AimPoint, TargetAvgVel/RangeRand(0.2, 4.0), self.Gun.Pos, 70)
			if aim then
				self.Ctrl.AnalogAim = Vector(1,0):RadRotate(aim+aimError+RangeRand(-0.018, 0.018))
			else
				break	-- target out of range
			end
			
			if ShootTimer:IsPastSimMS(aimTime) then
				self.fire = true
				aimError = aimError * RangeRand(0.95, 0.99)
			else
				self.fire = false
			end
		else
			self.deviceState = ACrab.POINTING
			self.fire = false
			
			ShootTimer:Reset()
			if self.FirearmIsEmpty then
				self:ReloadFirearm()
				aimTime = RangeRand(400, 700)
				aimError = RangeRand(-0.14, 0.14)
			else
				break -- no firearm avaliable
			end
		end
		
		coroutine.yield()	-- wait until next frame
		
		if not MovableMan:IsActor(self.Target) then
			break
		end
	end
	
	return true
end
]]
