
dofile("Base.rte/Constants.lua")
require("Actors/AI/PID")

function Create(self)
	self.GunTimer = {Timer(), Timer()}
	self.GunTimer[1]:SetSimTimeLimitMS(83)
	self.GunTimer[2]:SetSimTimeLimitMS(83)
	
	self.TargetLostTimer = {Timer(), Timer()}
	self.Target = {}
	self.TargetOffset = {}
	self.gunAng = {3.24, -0.1}
	self.limitHi = {4.08, 6.18}
	self.limitLo = {3.24, 5.34}
	self.maxAmmo = 30
	self.ammo = {self.maxAmmo, self.maxAmmo}
	self.Barrels = {}
	
	function self:LookForTargets(i)
		local Trace
		if self.Target[i] then
			Trace = Vector(FrameMan.PlayerScreenWidth*0.6, 0):RadRotate(self.gunAng[i]+RangeRand(-0.15, 0.15))
		else
			Trace = Vector(FrameMan.PlayerScreenWidth*0.6, 0):RadRotate(self.RotAngle+RangeRand(self.limitLo[i], self.limitHi[i]))
		end
		
		local ID = SceneMan:CastMORay(self.Barrels[i].Pos, Trace, self.ID, 5, rte.grassID, false, 6)
		if ID < 255 then
			return MovableMan:GetMOFromID(ID)
		end
		
		return nil
	end
	
	---------------- AI variables start ----------------
	self.StuckTimer = Timer()
	self.HatchTimer = Timer()
	
	self.PlayerInterferedTimer = Timer()
	self.PlayerInterferedTimer:SetSimTimeLimitMS(500)
	
	self.LZpos = Vector(self.Pos.X, 100)
	self.Controller = self:GetController()
	self.LastAIMode = Actor.AIMODE_NONE
	self.hoverAlt = self.Radius * 2.5	-- the drop ship tries to hover this many pixels above the ground
	
	if self.AIMode == Actor.AIMODE_DELIVER and self:IsInventoryEmpty() then
		self.AIMode = Actor.AIMODE_STAY	-- stop the craft from returning to orbit immediately
	end
	
	-- The controllers
	self.XposPID = RegulatorPID:New{p=0.1, i=0.00001, d=1.5, filter_leak=0.8}
	self.YposPD = RegulatorPID:New{p=0.1, d=2.5, filter_leak=0.8, integral=self.LZpos.Y}
	
	---------------- AI variables end ----------------
end

function Destroy(self)
	for j = 1, 2 do
		if MovableMan:IsParticle(self.Barrels[j]) then
			self.Barrels[j].ToDelete = true
		end
	end
end

function Update(self)
	for j = 1, 2 do
		-- Validate barrels
		if not MovableMan:IsParticle(self.Barrels[j]) or self.Barrels[j].PresetName ~= "MG Barrel" then
			self.Barrels[j] = CreateMOSRotating("MG Barrel", "EDV.rte")
			MovableMan:AddParticle(self.Barrels[j])
		end
		
		self.Barrels[j].RotAngle = self.gunAng[j]
	end
	
	-- Move barrels
	self.Barrels[1].Pos = self.Pos + self:RotateOffset(Vector(-52,-14))
	self.Barrels[2].Pos = self.Pos + self:RotateOffset(Vector(52,-14))
	
	for i = 1, 2 do
		local FoundMO = self:LookForTargets(i)
		if FoundMO then
			if self.Target[i] and MovableMan:IsActor(self.Target[i]) and FoundMO.RootID == self.Target[i].RootID then	-- found the same target
				self.TargetLostTimer[i]:Reset()
				self.TargetOffset[i] = (self.TargetOffset[i] + SceneMan:ShortestDistance(self.Target[i].Pos, SceneMan:GetLastRayHitPos(), false))/2
			else
				local MO = MovableMan:GetMOFromID(FoundMO.RootID)
				if MovableMan:IsActor(MO) then
					if MO.Team == self.Team then
						self.Target[i] = nil	-- stop shooting
					else
						if MO.ClassName == "AHuman" then
							MO = ToAHuman(MO)
						elseif MO.ClassName == "ACrab" then
							MO = ToACrab(MO)
						elseif MO.ClassName == "ACRocket" then
							MO = ToACRocket(MO)
						elseif MO.ClassName == "ACDropShip" then
							MO = ToACDropShip(MO)
						elseif MO.ClassName == "ADoor" then
							MO = ToADoor(MO)
						elseif MO.ClassName == "Actor" then
							MO = ToActor(MO)
						else
							MO = nil
						end
						
						if MO then
							self.TargetLostTimer[i]:Reset()
							if not self.Target[i] then
								self.Target[i] = MO
								self.TargetOffset[i] = SceneMan:ShortestDistance(self.Target[i].Pos, SceneMan:GetLastRayHitPos(), false)	-- this it the distance vector from the target center to the point we hit with our ray
								self.TargetLostTimer[i]:SetSimTimeLimitMS(1000)
							end
						end
					end
				end
			end
		else
			if self.Target[i] then
				if MovableMan:IsActor(self.Target[i]) then
					if self.TargetLostTimer[i]:IsPastSimTimeLimit() then
						self.Target[i] = nil	-- the target has been out of sight for a too long, ignore it
					end
				else
					self.Target[i] = nil
				end
			end
		end
		
		-- Track targets
		if self.Target[i] and MovableMan:IsActor(self.Target[i]) then
			local targetAng = SceneMan:ShortestDistance(self.Barrels[i].Pos, self.Target[i].Pos+self.TargetOffset[i], false).AbsRadAngle
			if targetAng > 6.28 then
				targetAng = targetAng - 6.28
			elseif targetAng < 0 then
				targetAng = targetAng + 6.28
			end
			
			local rotDiff = targetAng - self.gunAng[i]
			local tmp = self.gunAng[i] - targetAng
			if math.abs(tmp) < math.abs(rotDiff) then
				rotDiff = tmp
			end
			
			if math.abs(rotDiff) > 0.01 then
				local rot = 0
				if rotDiff > 0 then
					rot = math.min(0.02, rotDiff)
				else
					rot = math.max(-0.02, rotDiff)
				end
				
				self.gunAng[i] = math.min(math.max(self.gunAng[i] + rot, self.RotAngle + self.limitLo[i]), self.RotAngle + self.limitHi[i])
			end
			
			-- Fire
			if math.abs(rotDiff) < 0.2 and self.GunTimer[i]:IsPastSimTimeLimit() then
				self.GunTimer[i]:Reset()
				
				local bulletType = "Dropship MG"
				if self.ammo[i] % 4 == 0 then
					bulletType = "Dropship MG Tracer"
				end
				
				local Emitter = CreateAEmitter(bulletType, "EDV.rte")
				if Emitter then
					Emitter.RotAngle = self.gunAng[i]
					Emitter.Pos = self.Barrels[i].Pos + Vector(25, 0):RadRotate(Emitter.RotAngle)
					Emitter.Vel = self.Vel
					MovableMan:AddParticle(Emitter)
				end
				
				local Casing = CreateMOSParticle("Dropship MG Casing", "EDV.rte")
				if Casing then
					Casing.Pos = self.Barrels[i].Pos + Vector(15, 0):RadRotate(Emitter.RotAngle)
					Casing.RotAngle = self.gunAng[i]
					Casing.AngularVel = 2 + math.random() * 10
					Casing.Vel = self.Vel * (1.2 - math.random()*0.4)
					Casing.Vel.X = self.Vel.X + (i - 1.5) * 5
					MovableMan:AddParticle(Casing)
				end
				
				self.ammo[i] = self.ammo[i] - 1
				if self.ammo[i] < 1 then
					self.GunTimer[i]:SetSimTimeLimitMS(3000)
					self.ammo[i] = self.maxAmmo
				elseif self.ammo[i] == self.maxAmmo - 1 then
					self.GunTimer[i]:SetSimTimeLimitMS(83)
				end
			end
		else 	-- Reset barrel
			local rotDiff = (self.RotAngle + (self.limitLo[i] + self.limitHi[i])/2) - self.gunAng[i]
			if math.abs(rotDiff) > 0.1 then
				if rotDiff > 0 then
					rotDiff = math.min(0.01, rotDiff)
				else
					rotDiff = math.max(-0.01, rotDiff)
				end
				
				self.gunAng[i] = math.min(math.max(self.gunAng[i] + rotDiff, self.RotAngle + self.limitLo[i]), self.RotAngle + self.limitHi[i])
			end
		end
	end
end

function UpdateAI(self)
	if self.PlayerInterferedTimer:IsPastSimTimeLimit() then
		local FuturePos = self.Pos + self.Vel*20
		
		-- Make sure FuturePos is inside the scene
		if FuturePos.X > SceneMan.SceneWidth then
			if SceneMan.SceneWrapsX then
				FuturePos.X = FuturePos.X - SceneMan.SceneWidth
			else
				FuturePos.X = SceneMan.SceneWidth - self.Radius
			end
		elseif FuturePos.X < 0 then
			if SceneMan.SceneWrapsX then
				FuturePos.X = FuturePos.X + SceneMan.SceneWidth
			else
				FuturePos.X = self.Radius
			end
		end
		
		local Dist = SceneMan:ShortestDistance(FuturePos, self.LZpos, false)
		if math.abs(Dist.X) > 100 then
			if self.DeliveryState == ACraft.LAUNCH then
				self.LZpos.X = FuturePos.X
			else
				self.LZpos = SceneMan:MovePointToGround(FuturePos, self.hoverAlt, 5)
			end
		end
	end
	
	self.PlayerInterferedTimer:Reset()
	
	if self.AIMode ~= self.LastAIMode then
		self.LastAIMode = self.AIMode

		if self.AIMode == Actor.AIMODE_RETURN then
			self.DeliveryState = ACraft.LAUNCH
			self.LZpos = Vector(self.Pos.X, -500)	-- Go to orbit
		else -- Actor.AIMODE_STAY and Actor.AIMODE_DELIVER
			local FuturePos = self.Pos + self.Vel*20
			
			-- Make sure FuturePos is inside the scene
			if FuturePos.X > SceneMan.SceneWidth then
				if SceneMan.SceneWrapsX then
					FuturePos.X = FuturePos.X - SceneMan.SceneWidth
				else
					FuturePos.X = SceneMan.SceneWidth - self.Radius
				end
			elseif FuturePos.X < 0 then
				if SceneMan.SceneWrapsX then
					FuturePos.X = FuturePos.X + SceneMan.SceneWidth
				else
					FuturePos.X = self.Radius
				end
			end
			
			self.LZpos = SceneMan:MovePointToGround(FuturePos, self.hoverAlt, 5)
			self.DeliveryState = ACraft.FALL
		end
	end
	
	-- Control right/left movement
	local dist = SceneMan:ShortestDistance(self.Pos+self.Vel*10, self.LZpos, false).X
	local change = self.XposPID:Update(dist, 0)
	if change > 2 then
		self.Controller.AnalogMove = Vector(change/30, 0)
	elseif change < -2 then
		self.Controller.AnalogMove = Vector(change/30, 0)
	end

	-- Control up/down movement
	change = self.YposPD:Update(self.Vel.Y-(self.LZpos.Y-self.Pos.Y), 0)
	if change > 1 then
		self.AltitudeMoveState = ACraft.ASCEND
	elseif change < -1 then
		self.AltitudeMoveState = ACraft.DESCEND
	end

	-- Delivery Sequence logic
	if self.DeliveryState == ACraft.FALL then
		-- Don't descend if we have nothing to deliver
		if self:IsInventoryEmpty() then
			if self.AIMode ~= Actor.AIMODE_STAY then
				self.DeliveryState = ACraft.LAUNCH
				self.HatchTimer:Reset()
				self.LZpos.Y = -500	-- Go to orbit
			end
		else
			dist = SceneMan:ShortestDistance(self.Pos, self.LZpos, false).Magnitude
			if dist < 20 and math.abs(change) < 2 and self.Vel.Magnitude < 4 then	-- If we passed the hover check, start unloading
				if self.AIMode ~= Actor.AIMODE_STAY then
					self.DeliveryState = ACraft.UNLOAD
					self.HatchTimer:Reset()
				end
			else	-- Check for something in the way of our descent, and hover to the side to avoid it
				self.search = not self.search	-- Search every second update
				if self.search then
					local obstID = self:DetectObstacle(self.Radius * 7)
					if obstID > 0 then
						local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(obstID))
						if MO.ClassName == "ACDropShip" or MO.ClassName == "ACRocket" then
							self.LZpos.X = self.LZpos.X + 15
							self.AltitudeMoveState = ACraft.HOVER
						end
					end
				else
					local Free = Vector()
					local Start = self.Pos + Vector(self.Radius, 0)
					local Trace = self.Vel * (self.Radius/4) + Vector(0,50)
					if PosRand() < 0.5 then
						Start.X = Start.X - self.Diameter
					end
					
					if SceneMan:CastStrengthRay(Start, Trace, 0, Free, 4, 0, true) then
						self.LZpos.X = self.Pos.X
						self.LZpos.Y = Free.Y - self.hoverAlt
					end
				end
			end
		end
	elseif self.DeliveryState == ACraft.UNLOAD then
		if self.HatchTimer:IsPastSimMS(1000) then	-- Start unloading if there's something to unload
			self.HatchTimer:Reset()
			
			-- Randomly choose a direction to be going when unloading
			if PosRand() > 0.5 then
				self.LZpos.X = self.LZpos.X - 50
			else
				self.LZpos.X = self.LZpos.X + 50
			end
			
			if SceneMan.SceneWrapsX then
				if self.LZpos.X > SceneMan.SceneWidth then
					self.LZpos.X = self.LZpos.X - SceneMan.SceneWidth
				elseif self.LZpos.X < 0 then
					self.LZpos.X = self.LZpos.X + SceneMan.SceneWidth
				end
			end
			
			self.LZpos = SceneMan:MovePointToGround(Vector(self.LZpos.X, self.LZpos.Y - self.Diameter), self.hoverAlt, 6)
			self.Controller:SetState(Controller.PRESS_FACEBUTTON, true)
			self.DeliveryState = ACraft.FALL
		end
	elseif self.DeliveryState == ACraft.LAUNCH then
		if self.HatchTimer:IsPastSimMS(1000) then
			self.HatchTimer:Reset()
			self:CloseHatch()
		end

		-- Check for something in the way of our ascent, and hover to the side to avoid it
		self.search = not self.search	-- Search every second update
		if self.search then
			local obstID = self:DetectObstacle(self.Radius * 7)
			if obstID > 0 then
				local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(obstID))
				if MO.ClassName == "ACDropShip" or MO.ClassName == "ACRocket" then
					self.LZpos.X = self.LZpos.X - 15
					self.AltitudeMoveState = ACraft.HOVER
				end
			end
		end
	end

	-- If we are hopelessly stuck, self destruct
	if self.Vel.Magnitude > 1 then
		self.StuckTimer:Reset()
	elseif self.StuckTimer:IsPastSimMS(20000) then
		self:GibThis()
	end

	-- Input translation
	if self.AltitudeMoveState == ACraft.ASCEND then
		self.Controller:SetState(Controller.MOVE_UP, true)
	elseif self.AltitudeMoveState == ACraft.DESCEND then
		self.Controller:SetState(Controller.MOVE_DOWN, true)
	else
		self.Controller:SetState(Controller.MOVE_UP, false)
		self.Controller:SetState(Controller.MOVE_DOWN, false)
	end
end
