
FlyingDummyAI = {}

function Create(self)
	---------------- AI variables start ----------------
	self.Ctrl = self:GetController()
	
	self.lateralMoveState = Actor.LAT_STILL
	self.jumpState = AHuman.NOTJUMPING
	
	self.deviceState = AHuman.STILL
	self.lastAIMode = Actor.AIMODE_NONE
	self.teamBlockState = Actor.NOTBLOCKED
	self.SentryFacing = self.HFlipped
	self.fire = false
	
	self.AirTimer = Timer()
	self.PickUpTimer = Timer()
	self.ReloadTimer = Timer()
	self.BlockedTimer = Timer()
	self.TargetLostTimer = Timer()
	
	self.JumpTimer = Timer()
	self.JumpTimer:SetSimTimeLimitMS(math.floor(RangeRand(1000, 2000)))
	
	self.PlayerInterferedTimer = Timer()
	self.PlayerInterferedTimer:SetSimTimeLimitMS(500)
	
	self.AlarmTimer = Timer()
	self.AlarmTimer:SetSimTimeLimitMS(500)
	
	-- check if this team is controlled by a human
	local Activ = ActivityMan:GetActivity()
	for player = Activity.PLAYER_1, Activity.MAXPLAYERCOUNT - 1 do
		if Activ:PlayerActive(player) and Activ:PlayerHuman(player) then
			if self.Team == Activ:GetTeamOfPlayer(player) then
				self.isPlayerOwned = true
				break
			end
		end
	end
	
	function self:LookForTargets()
		local Origin
		local viewAng
		
		if self.deviceState == AHuman.AIMING and self.FirearmIsReady then
			Origin = ToHeldDevice(self.EquippedItem).MuzzlePos
			viewAng = 0.1
		elseif self.deviceState == AHuman.POINTING then
			Origin = self.EyePos
			viewAng = 0.3
		elseif self.deviceState == AHuman.THROWING then
			Origin = self.EyePos
			viewAng = 0.4
		else
			Origin = self.EyePos
			viewAng = 0.7
		end
		
		local viewLen = SceneMan:ShortestDistance(self.EyePos, self.ViewPoint, false).Magnitude + FrameMan.PlayerScreenWidth*0.5
		local Trace = Vector(viewLen, 0):RadRotate(viewAng*NormalRand()+self:GetAimAngle(true))
		local ID = SceneMan:CastMORay(Origin, Trace, self.ID, grassID, false, 5)
		
		if ID < 255 then
			local Point = SceneMan:GetLastRayHitPos()
			if not SceneMan:IsUnseen(Point.X, Point.Y, self.Team) then
				return MovableMan:GetMOFromID(ID)
			end
		end
		
		return nil
	end
	
	-- functions that create behaviors. the default behaviors are stored in the HumanAI table. store your custom behaviors in a table to avoid name conflicts between mods.
	function self:CreateSentryBehavior()
		if not self.FirearmIsReady and not self.ThrowableIsReady then
			return
		end
		
		self.NextBehavior = coroutine.create(HumanAI.Sentry)	-- replace "HumanAI.Sentry" with the function name of your own sentry behavior
		self.NextCleanup = nil
		self.NextBehaviorName = "Sentry"
	end
	
	function self:CreatePatrolBehavior()
		self.NextBehavior = coroutine.create(HumanAI.Patrol)
		self.NextCleanup = nil
		self.NextBehaviorName = "Patrol"
	end
	
	function self:CreateGoldDigBehavior()
		self.NextBehavior = coroutine.create(HumanAI.GoldDig)
		self.NextCleanup = nil
		self.NextBehaviorName = "GoldDig"
	end

	function self:CreateBrainSearchBehavior()
		self.NextBehavior = coroutine.create(HumanAI.BrainSearch)
		self.NextCleanup = nil
		self.NextBehaviorName = "BrainSearch"
	end
	
	function self:CreateGetToolBehavior()
		self.NextBehavior = coroutine.create(HumanAI.ToolSearch)
		self.NextCleanup = nil
		self.NextBehaviorName = "ToolSearch"
	end
	
	function self:CreateGetWeaponBehavior()
		self.NextBehavior = coroutine.create(HumanAI.WeaponSearch)
		self.NextCleanup = nil
		self.NextBehaviorName = "WeaponSearch"
	end
	
	function self:CreateGoToBehavior()
		self.NextGoTo = coroutine.create(HumanAI.GoToWpt)
		self.NextGoToCleanup = function(self)
			self.lateralMoveState = Actor.LAT_STILL
			self.deviceState = AHuman.STILL
			self.crawling = false
			self.jump = false
			self.fire = false
			self:EquipFirearm(true)
			self:UpdateMovePath()
		end
		self.NextGoToName = "GoToWpt"
	end
		
	function self:CreateMoveAroundBehavior()
		self.NextGoTo = coroutine.create(HumanAI.MoveAroundActor)
		self.NextGoToCleanup = function(self)
			self.lateralMoveState = Actor.LAT_STILL
			self.jump = false
		end
		self.NextGoToName = "MoveAroundActor"
	end
	
	function self:CreateAttackBehavior()
		if self:EquipFirearm(true) then
			self.NextBehavior = coroutine.create(HumanAI.ShootTarget)
			self.NextBehaviorName = "ShootTarget"
		elseif self:EquipThrowable(true) then
			self.NextBehavior = coroutine.create(HumanAI.ThrowTarget)
			self.NextBehaviorName = "ThrowTarget"
		elseif self:EquipDiggingTool(true) then
			self.NextBehavior = coroutine.create(HumanAI.DigTarget)
			self.NextBehaviorName = "DigTarget"
		else
			return	-- unarmed	TODO: start run away behavior here?
		end
		
		self.NextCleanup = function(self)
			self.fire = false
			self.Target = nil
			self.deviceState = AHuman.STILL
		end
	end
	
	function self:CreateJumpBehavior()
		self.NextGoTo = coroutine.create(FlyingDummyAI.Jump)
		self.NextGoToCleanup = nil
		self.NextGoToName = "Jump"
	end
	
	-- the controller class
	PIDController = {}
	PIDController.mt = {__index = PIDController}
	function PIDController:New(p_in, i_in, d_in, last_in, integral_in)
		return setmetatable(
			{
				p = p_in,
				i = i_in,
				d = d_in,
				last = last_in,
				integral = integral_in or 0
			},
			PIDController.mt)
	end
	
	function PIDController:Update(input, target)
		local err = input - target
		local change = input - self.last
		self.last = input
		self.integral = self.integral + err
		self.integral = math.min(math.max(self.integral, -100), 100)
		return self.p*err + self.i*self.integral + self.d*change
	end
	
	-- the controllers
	self.XposPID = PIDController:New(0.04, 0.00001, 0.4, 0)
	self.YposPID = PIDController:New(0.05, 0.0001, 0.2, 0)
	
	---------------- AI variables end ----------------
end

function UpdateAI(self)	
	if self.PlayerInterferedTimer:IsPastSimTimeLimit() then
		self.Behavior = nil	-- remove the current behavior
		if self.BehaviorCleanup then
			self.BehaviorCleanup(self)	-- clean up after the current behavior
			self.BehaviorCleanup = nil
		end
		
		self.GoToBehavior = nil
		if self.GoToCleanup then
			self.GoToCleanup(self)
			self.GoToCleanup = nil
		end
		
		self.Target = nil
		self.PickupHD = nil
		self.BlockingActor = nil
		self.FollowingActor = nil
		
		self.fire = false
		self.jump = false
		self.crawling = false
		
		self.deviceState = AHuman.STILL
		self.lastAIMode = Actor.AIMODE_NONE
		self.teamBlockState = Actor.NOTBLOCKED
	end
	
	self.PlayerInterferedTimer:Reset()
	
	-- switch to the next behavior, if avaliable
	if self.NextBehavior then
		if self.BehaviorCleanup then
			self.BehaviorCleanup(self)
		end
		
		self.Behavior = self.NextBehavior
		self.BehaviorCleanup = self.NextCleanup
		self.BehaviorName = self.NextBehaviorName
		self.NextBehavior = nil
	end
	
	-- switch to the next GoTo behavior, if avaliable
	if self.NextGoTo then
		if self.GoToCleanup then
			self.GoToCleanup(self)
		end
		
		self.GoToBehavior = self.NextGoTo
		self.GoToCleanup = self.NextGoToCleanup
		self.NextGoTo = nil
	end
	
	-- check if the legs reach the ground
	if self.AirTimer:IsPastSimMS(250) then
		self.AirTimer:Reset()
		
		if -1 < SceneMan:CastObstacleRay(self.Pos, Vector(0, self.Height/4), Vector(), Vector(), self.ID, grassID, 3) then
			self.flying = false
		else
			self.flying = true
		end
	end
	
	-- check if the AI mode has changed or if we need a new behavior
	if (self.AIMode ~= self.lastAIMode or (not self.Behavior and not self.GoToBehavior)) and not self.flying then
		self.Behavior = nil
		if self.BehaviorCleanup then
			self.BehaviorCleanup(self)	-- stop the current behavior
			self.BehaviorCleanup = nil
		end
		
		self.GoToBehavior = nil
		if self.GoToCleanup then
			self.GoToCleanup(self)
			self.GoToCleanup = nil
		end
		
		-- select a new behavior based on AI mode
		if self.AIMode == Actor.AIMODE_GOTO then
			self:CreateGoToBehavior()
		elseif self.AIMode == Actor.AIMODE_BRAINHUNT then
			self:CreateBrainSearchBehavior()
		elseif self.AIMode == Actor.AIMODE_GOLDDIG then
			self:CreateGoldDigBehavior()
		elseif self.AIMode == Actor.AIMODE_PATROL then
			self:CreatePatrolBehavior()
		else
			if self.AIMode ~= self.lastAIMode and self.AIMode == Actor.AIMODE_SENTRY then
				self.SentryFacing = self.HFlipped	-- store the direction in which we should be looking
			end
			
			self:CreateSentryBehavior()
		end
		
		self.lastAIMode = self.AIMode
	end
	
	-- look for targets
	local FoundMO = self:LookForTargets()
	if FoundMO then
		if self.Target and MovableMan:IsActor(self.Target) and FoundMO.RootID == self.Target.RootID then	-- found the same target
			local Point = SceneMan:GetLastRayHitPos()
			self.TargetOffset = (self.TargetOffset + SceneMan:ShortestDistance(self.Target.Pos, Point, false))/2
			self.ReloadTimer:Reset()
			if not SceneMan:IsUnseen(Point.X, Point.Y, self.Team) then
				self.TargetLostTimer:Reset()
			end
		else
			local MO = MovableMan:GetMOFromID(FoundMO.RootID)
			if MovableMan:IsActor(MO) then
				if MO.Team == self.Team then
					self.Target = nil	-- stop shooting
					if MO.ClassName ~= "ADoor" and
						SceneMan:ShortestDistance(self.Pos, MO.Pos, false).Magnitude < self.Diameter + MO.Diameter
					then
						self.BlockingActor = ToActor(MO)
					end
				else
					if MO.ClassName == "AHuman" then
						MO = ToAHuman(MO)
					elseif MO.ClassName == "ACrab" then
						MO = ToACrab(MO)
					elseif MO.ClassName == "ACRocket" then
						MO = ToACRocket(MO)
					elseif MO.ClassName == "ACDropShip" then
						MO = ToACDropShip(MO)
					elseif MO.ClassName == "ADoor" then
						MO = ToADoor(MO)
					elseif MO.ClassName == "Actor" then
						MO = ToActor(MO)
					else
						MO = nil
					end
					
					if MO then
						self.ReloadTimer:Reset()
						
						if not self.Target then
							self.OldTargetPos = nil
							self.Target = MO
							self.TargetOffset = SceneMan:ShortestDistance(self.Target.Pos, SceneMan:GetLastRayHitPos(), false)	-- this it the distance vector from the target center to the point we hit with our ray
							self.TargetLostTimer:SetSimTimeLimitMS(1000)
							self:CreateAttackBehavior()
						--else
							-- TODO: check if this target have higher priority
						end
						
						local Point = self.Target.Pos + self.TargetOffset
						if not SceneMan:IsUnseen(Point.X, Point.Y, self.Team) then
							self.TargetLostTimer:Reset()
						end
					end
				end
			end
		end
	else
		if self.Target then
			if MovableMan:IsActor(self.Target) then
				if self.TargetLostTimer:IsPastSimTimeLimit() then
					self.OldTargetPos = self.Target.Pos
					self.Target = nil	-- the target has been out of sight for a too long, ignore it
				end
			else
				self.Target = nil
			end
		end
		
		if self.ReloadTimer:IsPastSimMS(8000) then	-- check if we need to reload
			self.ReloadTimer:Reset()
			if ToAHuman(self).FirearmNeedsReload then
				self:ReloadFirearm()
			end
		end
	end
	
	-- check if it is possible to jump to the target
	if self.GoToBehavior and not self.Behavior and not self.flying and 
		 self.AIMode ~= Actor.AIMODE_GOLDDIG and self.JumpTimer:IsPastSimTimeLimit()
	then
		self.JumpTimer:Reset()
		
		local LZ = self:GetLastAIWaypoint()
		local Dist = SceneMan:ShortestDistance(LZ, self.Pos, false)
		if Dist.Magnitude > 100 then
			-- check for obstacles above
			if SceneMan:CastObstacleRay(self.Pos, Vector(0, -40), Vector(), Vector(), self.ID, grassID, 7) < 0 then
				self.GoToBehavior = nil
				if self.GoToCleanup then
					self.GoToCleanup(self)
					self.GoToCleanup = nil
				end
				
				self:CreateJumpBehavior()
				self.JumpTimer:SetSimTimeLimitMS(math.floor(RangeRand(6000, 7000)))
			end
		end
	end
	
	if self.GoToBehavior then -- run the move behavior and delete it if it returns true
		local msg, done = coroutine.resume(self.GoToBehavior, self)
		if not msg then
			ConsoleMan:PrintString(self.PresetName .. " GoToBehavior error:\n" .. done)	-- print the error message
			done = true
		end
		
		if done then
			self.GoToBehavior = nil
			if self.GoToCleanup then
				self.GoToCleanup(self)
				self.GoToCleanup = nil
			end
		end
	elseif self.flying and not self.jump and self.Vel.Y > 10 then
		self.Ctrl:SetState(Controller.BODY_JUMP, true)
	end
	
	-- run the selected behavior and delete it if it returns true
	if self.Behavior then
		local msg, done = coroutine.resume(self.Behavior, self)
		if not msg then
			ConsoleMan:PrintString(self.PresetName .. " behavior " .. self.BehaviorName .. " error:\n" .. done)	-- print the error message
			done = true
		end
		
		if done then
			self.Behavior = nil
			if self.BehaviorCleanup then
				self.BehaviorCleanup(self)
				self.BehaviorCleanup = nil
			end
			
			if not self.PickupHD and not self.NextBehavior and self.PickUpTimer:IsPastSimMS(500) then
				self.PickUpTimer:Reset()
				
				if not self:EquipFirearm(false) then
					self:CreateGetWeaponBehavior()
				elseif self.AIMode ~= Actor.AIMODE_SENTRY and not self:EquipDiggingTool(false) then
					self:CreateGetToolBehavior()
				end
			end
		end
	elseif self.Target and self.EquippedItem and ToHeldDevice(self.EquippedItem):IsTool() then	-- attack with digger
		-- TODO: move this to a separate behavior
		if MovableMan:IsActor(self.Target) then
			local Dist = SceneMan:ShortestDistance(ToHeldDevice(self.EquippedItem).MuzzlePos, self.Target.Pos, false)
			if Dist.Magnitude < 50 then
				self.Ctrl.AnalogAim = Dist.Normalized
				self.Ctrl:SetState(Controller.WEAPON_FIRE, true) -- fire digger
			end
		else
			self.Target = nil
		end
	elseif not self.PickupHD and not self.NextBehavior and self.PickUpTimer:IsPastSimMS(500) then
		self.PickUpTimer:Reset()
		
		if not self:EquipFirearm(false) then
			self:CreateGetWeaponBehavior()
		elseif self.AIMode ~= Actor.AIMODE_SENTRY and not self:EquipDiggingTool(false) then
			self:CreateGetToolBehavior()
		end
	end
	
	if self.PickupHD then	-- there is a HeldDevice we want to pick up
		if not MovableMan:IsDevice(self.PickupHD) or self.PickupHD.ID ~= self.PickupHD.RootID then
			self.PickupHD = nil	-- the HeldDevice has been destroyed or picked up
			self:ClearAIWaypoints()
			
			if self.PrevAIWaypoint or self.FollowingActor then				
				if self.FollowingActor and MovableMan:IsActor(self.FollowingActor) then	-- what if the old destination was a moving actor?
					self:AddAIMOWaypoint(self.FollowingActor)
					self:CreateGoToBehavior()	-- continue towards our old destination
				else
					self.FollowingActor = nil
					if self.PrevAIWaypoint then
						self:AddAISceneWaypoint(self.PrevAIWaypoint)
						self:CreateGoToBehavior()	-- continue towards our old destination
					end
				end
			end
		else
			if SceneMan:ShortestDistance(self.Pos, self.PickupHD.Pos, false).Magnitude < self.Height then
				self.Ctrl:SetState(Controller.WEAPON_PICKUP, true)
			end
		end
	end
	
	-- react to relevant AlarmEvents
	if self.AlarmTimer:IsPastSimTimeLimit() then
		self.AlarmPos = nil
		
		for Event in MovableMan.AlarmEvents do
			if Event.Team ~= self.Team then	-- caused by some other team's activites - alarming!
				local AlarmVec = SceneMan:ShortestDistance(self.EyePos, Event.ScenePos, false)	-- see how far away the alarm situation is
				
				-- only react if the alarm is within range and we are perceptive enough to hear it
				if AlarmVec.Largest <= Event.Range then	-- * self.Perceptiveness then
					-- if this is the same alarm location as last, then don't repeat the signal
					if not self.LastAlarmPos or SceneMan:ShortestDistance(self.LastAlarmPos, Event.ScenePos, false).Largest > 10 then
						-- now check if we have line of sight to the alarm point
						-- don't check all the way to the target, we are checking for no obstacles, and target will be an obstacle in itself
						if SceneMan:CastObstacleRay(self.EyePos, AlarmVec*0.9, Vector(), Vector(), self.ID, grassID, 7) < 0 then
							self.LastAlarmPos = self.AlarmPos
							self.AlarmPos = Vector(Event.ScenePos.X, Event.ScenePos.Y)
							self.AlarmTimer:Reset()
							break
						end
					end
				end
			end
		end
	elseif self.AlarmPos then	-- if alarmed, look at the alarming point until AlarmTimer has expired
		-- point in the direciton we heard the alarm come from if we're not already engaging a target
		if not (self.Target and
			 (self.deviceState == AHuman.AIMING or
			  self.deviceState == AHuman.FIRING or
			  self.deviceState == AHuman.THROWING))
		then
			-- look/point in the direction of alarm
			self.deviceState = AHuman.POINTING
			self.Ctrl.AnalogAim = SceneMan:ShortestDistance(self.EyePos, self.AlarmPos, false).Normalized
			self.lateralMoveState = Actor.LAT_STILL
			self.jump = false
		end
		
		-- if we're unarmed, hit the deck!
		if not self:EquipFirearm(true) and not self:EquipThrowable(true) and not self:EquipDiggingTool(true) then
			self.Ctrl:SetState(Controller.BODY_CROUCH, true)
			-- TODO: run away behavior
		end
	end
	
	if self.teamBlockState == Actor.IGNORINGBLOCK then
		if self.BlockedTimer:IsPastSimMS(10000) then
			self.teamBlockState = Actor.NOTBLOCKED
		end
	elseif self.teamBlockState == Actor.BLOCKED then	-- we are blocked by a teammate, stop
		self.lateralMoveState = Actor.LAT_STILL
		self.jump = false
		if self.BlockedTimer:IsPastSimMS(20000) then
			self.BlockedTimer:Reset()
			self.teamBlockState = Actor.IGNORINGBLOCK
		end
	else
		self.BlockedTimer:Reset()
	end
	
	-- controller states
	if self.fire then
		self.Ctrl:SetState(Controller.WEAPON_FIRE, true)
	end
		
	if self.deviceState == AHuman.AIMING then
		self.Ctrl:SetState(Controller.AIM_SHARP, true)
	end
	
	if self.JumpData then
		self.Vel = self.Vel + Vector(1.2,0):RadRotate(self.JumpData.ang)
		self.JumpData.acc = self.JumpData.acc - 1
		if self.JumpData.acc < 1 then
			self.JumpData = nil
		end
		
		if self.JumpThruster and MovableMan:ValidMO(self.JumpThruster) then
			self.JumpThruster.Pos = self.Pos + self:RotateOffset(Vector(-6, -1))
		end
	else
		if self.crawling then
			self.Ctrl:SetState(Controller.BODY_CROUCH, true)
		end
		
		if self.jump then
			if self.jumpState == AHuman.PREJUMP then
				self.jumpState = AHuman.UPJUMP
			elseif self.jumpState ~= AHuman.UPJUMP then
				self.jumpState = AHuman.PREJUMP
			end
			
			if self.JetTimeLeft < self.JetTimeTotal * 0.05 then	-- TODO: move this to gotowpt?
				self.refuel = true
			end
		else
			self.jumpState = AHuman.NOTJUMPING
		end
		
		if self.jumpState == AHuman.PREJUMP then
			self.Ctrl:SetState(Controller.BODY_JUMPSTART, true)
			self.Ctrl:SetState(Controller.BODY_JUMP, false)
		elseif self.jumpState == AHuman.UPJUMP then
			self.Ctrl:SetState(Controller.BODY_JUMPSTART, false)
			self.Ctrl:SetState(Controller.BODY_JUMP, true)
		end
		
		if self.lateralMoveState == Actor.LAT_LEFT then
			self.Ctrl:SetState(Controller.MOVE_LEFT, true)
			self.Ctrl:SetState(Controller.MOVE_RIGHT, false)
		elseif self.lateralMoveState == Actor.LAT_RIGHT then
			self.Ctrl:SetState(Controller.MOVE_LEFT, false)
			self.Ctrl:SetState(Controller.MOVE_RIGHT, true)
		end
	end
end


-----------------------
------ Behaviors ------
-----------------------

function FlyingDummyAI.Jump(self)
	while self.Vel.Magnitude > 2 do
		coroutine.yield()	-- wait until next frame
	end

	local Solutions = {}
	local LZ = self:GetLastAIWaypoint()
	
	--DebugIcon("Cross", LZ)
	
	local Dist = SceneMan:ShortestDistance(LZ, self.Pos, false)
	local SI2Pixels = FrameMan.PPM * TimerMan.DeltaTimeSecs -- used to convert m/s to pixels/frame
	local StopPos, LastPos, NewPos, Acc, Vel, minAng, maxAng, range
	
	if Dist.X > 0 then	-- LZ is to the left
		self.HFlipped = true
		minAng = 1.64
		if LZ.Y < self.Pos.Y then	-- LZ above us
			maxAng = 2.34
		else
			maxAng = 2.74
		end
	else	-- LZ is to the right
		self.HFlipped = false
		maxAng = 1.5
		if LZ.Y < self.Pos.Y then	-- LZ above us
			minAng = 0.8
		else
			minAng = 0.4
		end
	end
	
	for ang = minAng, maxAng, 0.05 do
		for acc = 28, 8, -8 do
			local Acc = Vector(1.2,0):RadRotate(ang) + SceneMan.GlobalAcc / (FrameMan.PPM / SI2Pixels) -- jetpack and gravity acceleration in pixels/frame/frame
			
			NewPos = Vector(self.Pos.X, self.Pos.Y)
			Vel = Vector(self.Vel.X, self.Vel.Y)
			
			local fail = false
			for t = 1, acc do
				Vel = Vel + Acc
				NewPos = NewPos + Vel * SI2Pixels
				
				if SceneMan:GetTerrMatter(NewPos.X, NewPos.Y) ~= 0 then
					fail = true
					break
				end
			end
			
			if not fail then
				coroutine.yield()	-- wait until next frame
				
				LastPos = Vector(NewPos.X, NewPos.Y)
				Acc = SceneMan.GlobalAcc / (FrameMan.PPM / SI2Pixels) -- acceleration in pixels/frame/frame
				StopPos = Vector(NewPos.X, NewPos.Y)	-- this is the point where acceleration will stop
				
				for t = 1, 60, 4 do
					NewPos = StopPos + Acc * ((t^2)/2) / SI2Pixels + Vel * t					
					if NewPos.Y < SceneMan.SceneHeight and SceneMan:GetTerrMatter(NewPos.X, NewPos.Y) == 0 then
						LastPos:SetXY(NewPos.X, NewPos.Y)
						--DebugIcon("Ring 1 Long", LastPos)
					else
						break
					end
				end
				
				Dist = SceneMan:ShortestDistance(LastPos, LZ, false)
				range = Dist.Magnitude
				if range < 600 then
					table.insert(Solutions, {StopPos=Vector(StopPos.X, StopPos.Y), StopVel=Vector(Vel.X, Vel.Y), acc=acc, ang=ang, range=range})
				end
			end
			
			coroutine.yield()	-- wait until next frame
		end
	end
	
	if #Solutions > 0 then
		table.sort(Solutions, function(A,B) return A.range < B.range end)
		while #Solutions > 5 do
			table.remove(Solutions)	-- limit to five trials
		end
		
		local BestSls = {}
		
		-- test again with a higher resolution, starting with the most likely candidate
		for _,Sl in pairs(Solutions) do
			coroutine.yield()	-- wait until next frame
			
			LastPos:SetXY(0,0)
			Acc = SceneMan.GlobalAcc / (FrameMan.PPM / SI2Pixels) -- acceleration in pixels/frame/frame
			StopPos = Sl.StopPos
			Vel = Sl.StopVel
			
			for t = 1, 60 do
				NewPos = StopPos + Acc * ((t^2)/2) / SI2Pixels + Vel * t					
				if NewPos.Y < SceneMan.SceneHeight and
					 SceneMan:GetTerrMatter(NewPos.X, NewPos.Y) == 0 and
					 SceneMan:GetTerrMatter(NewPos.X + RangeRand(-5,5), NewPos.Y + RangeRand(-5,5)) == 0
				then
					LastPos:SetXY(NewPos.X, NewPos.Y)
					--DebugIcon("Ring 2 Long", LastPos)
				else
					break
				end
			end
			
			Dist = SceneMan:ShortestDistance(LastPos, LZ, false)
			range = Dist.Magnitude
			
			if SceneMan:CastObstacleRay(LastPos, Dist, Vector(), Vector(), self.ID, grassID, 4) < 0 then
				range = range / 2
			end
			
			table.insert(BestSls, {acc=Sl.acc, ang=Sl.ang, range=range})
		end
		
		if #BestSls > 0 then
			table.sort(BestSls, function(A,B) return A.range < B.range end)	-- sort by distance to the LZ
			
			self.JumpData = BestSls[1]	-- pick the best of the best
			self.Pos = self.Pos + Vector(0, -3)	-- disloge the actors feet
			
			-- acceleration last 32, 24, 16 or 8 frames
			if self.JumpData.acc < 16 then
				self.JumpThruster = CreateAEmitter("Jump Thruster 1", "EDV.rte")
			elseif self.JumpData.acc < 24 then
				self.JumpThruster = CreateAEmitter("Jump Thruster 2", "EDV.rte")
			elseif self.JumpData.acc < 32 then
				self.JumpThruster = CreateAEmitter("Jump Thruster 3", "EDV.rte")
			else
				self.JumpThruster = CreateAEmitter("Jump Thruster 4", "EDV.rte")
			end
			
			if self.JumpThruster then
				self.JumpThruster.Pos = self.Pos + self:RotateOffset(Vector(-6, -1))
				self.JumpThruster.RotAngle = self.JumpData.ang + 3.14
				MovableMan:AddParticle(self.JumpThruster)
			end
		else
			self:CreateGoToBehavior()
		end
	else
		self:CreateGoToBehavior()
	end
	
	return true
end
