--[[	A special thank you to DarlosD9 and Darkstorm Technologies for this list of key numbers.

	A - 1		S - 19		Numpad 0 - 37		ins  - 76
	B - 2		T - 20		Numpad 1 - 38		del  - 77
	C - 3		U - 21		Numpad 2 - 39		home - 78
	D - 4		V - 22		Numpad 3 - 40		end  - 79
	E - 5		W - 23		Numpad 4 - 41		page up - 80
	F - 6		X - 24		Numpad 5 - 42		page down - 81
	G - 7		Y - 25		Numpad 6 - 43		left arrow - 82
	H - 8		Z - 26		Numpad 7 - 44		right arrow - 83
	I - 9		0 - 27		Numpad 8 - 45		up arrow - 84
	J - 10		1 - 28		Numpad 9 - 46		down arrow - 85
	K - 11		2 - 29		- - 61			numpad / - 86
	L - 12		3 - 30		= - 62			numpad * - 87
	M - 13		4 - 31		{ - 65			numpad - - 88
	N - 14		5 - 32		} - 66			numpad + - 89
	O - 15		6 - 33		" - 69			numpad . - 90
	P - 16		7 - 34		\ - 70			;  - 105
	Q - 17		8 - 35		, - 72
	R - 18		9 - 36		. - 73
]]

require("Actors/AI/PID")

function Create(self)
	---------------- AI variables start ----------------
	self.StuckTimer = Timer()
	self.PlayerInterferedTimer = Timer()
	self.PlayerInterferedTimer:SetSimTimeLimitMS(500)
	
	self.LZpos = Vector(self.Pos.X, 100)
	self.LastAIMode = Actor.AIMODE_NONE
	
	if self.AIMode ~= Actor.AIMODE_GOTO then
		self.AIMode = Actor.AIMODE_STAY	-- stop the craft from returning to orbit
	end
	
	-- The controllers
	self.XposPID = RegulatorPID:New{p=0.06, i=0.00001, d=0.7, filter_leak=0.8}
	self.YposPD = RegulatorPID:New{p=0.1, d=2.5, filter_leak=0.8, integral=self.LZpos.Y}
	---------------- AI variables end ----------------
	
	function LauncherKey()
		return (UInputMan:KeyHeld(24) or UInputMan:KeyHeld(26))	-- X or Z, replace with a number form the list above
	end
	
	function Offset(Origin, ang, dist)
		if Origin then
			return Vector(Origin.X, Origin.Y) + Vector(dist, 0):RadRotate(ang)
		end
		return Vector(dist, 0):RadRotate(ang)
	end
	
	-- [-pi .. pi]
	function NormalizeAngle(angle)	
		return angle - 6.28319 * math.floor(angle/6.28319+0.5)
	end
	
	-- [0 .. pi*2]
	function LimitAngle(angle)
		if angle >= 0 then
			return angle - math.floor(angle/6.28319)*6.28319
		else
			angle = angle/6.28319
			return -6.28319 * math.floor(angle) + angle*6.28319
		end
	end
	
	self.Barrel = {}
	self.Offset = {}
	self.AimDots = {}
	self.maxAmmo = 28
	self.ammo = self.maxAmmo
	self.GunTimer = Timer()
	self.GrenadeTimer = Timer()
	self.RechargeTimer = Timer()
	self.Ctrl = self:GetController()
	self.AimPos = Vector()
	self.canFire = {true, true, true}
	self.grenades = 0
	self.laserBurst = 120
	self.laserNormal = 250
	self.laserRof = self.laserNormal
	self.laserStartRoT = 640
	self.laserRoT = self.laserStartRoT		-- Reload time
	self.aim = 0
	self.twoFrames = math.ceil(TimerMan.DeltaTimeSecs * 2000)
	
	-- Remove any inventory
	if not self:IsInventoryEmpty() then
		-- Create an empty gunship
		local Act = CreateACDropShip("Dummy Gunship", "EDV.rte")
		if Act then
			Act.Pos = self.Pos
			Act.Vel = self.Vel
			Act.RotAngle = self.RotAngle
			Act.Team = self.Team
			MovableMan:AddActor(Act)
		end
		
		self.Pos.Y = -99
		self.Vel = Vector(0, -100)
		self.BarrelName = {"No Barrel", "No Barrel", "No Barrel"}
	else
		self.BarrelName = {"Laser Barrel", "Laser Barrel", "Launcher Barrel"}
	end
end

function Destroy(self)
	if self.Health < 1 then
		for i = 1, #self.Barrel do
			if MovableMan:ValidMO(self.Barrel[i]) and self.Barrel[i].PresetName == self.BarrelName[i] then
				self.Barrel[i]:GibThis()
			end
		end
	else
		for i = 1, #self.Barrel do
			if MovableMan:ValidMO(self.Barrel[i]) and self.Barrel[i].PresetName == self.BarrelName[i] then
				self.Barrel[i].ToDelete = true
			end
		end		
	end
	
	for i = 1, 7 do
		if MovableMan:ValidMO(self.AimDots[i]) and self.AimDots[i].PresetName == "Aim" then
			self.AimDots[i].ToDelete = true
		end
	end
	
	if MovableMan:ValidMO(self.AmmoIcon) and self.AmmoIcon.PresetName == "Ammo Counter" then
		self.AmmoIcon.ToDelete = true
	end
end

function Update(self)
	local gunAim = 0
	
	if self.ammo < self.maxAmmo and self.RechargeTimer:IsPastSimMS(self.laserRoT) then
		self.RechargeTimer:Reset()
		self.laserRoT = math.max(self.laserRoT * 0.72, self.twoFrames)
		self.ammo = self.ammo + 1
	end
	
	-- Air resistance
	if math.abs(self.Vel.X) > 10 then
		self.Vel.X = self.Vel.X * 0.997
	end
	
	-- Update weapon objects
	self.Offset[1] = self.Pos + self:RotateOffset(Vector(-80, 0))
	self.Offset[2] = self.Pos + self:RotateOffset(Vector(80, 0))
	self.Offset[3] = self.Pos + self:RotateOffset(Vector(0, -40))
	
	for i = 1, #self.BarrelName do
		if MovableMan:ValidMO(self.Barrel[i]) and self.Barrel[i].PresetName == self.BarrelName[i] then
			self.Barrel[i].Pos = self.Offset[i]
			self.Barrel[i].Age = 1
			self.Barrel[i].AngularVel = 0
			self.canFire[i] = true
		elseif self.Barrel[i] == nil then
			self.Barrel[i] = CreateMOSRotating(self.BarrelName[i], "EDV.rte")
			self.Barrel[i].Pos = self.Offset[i]
			self.Barrel[i].RotAngle = SceneMan:ShortestDistance(self.Barrel[i].Pos, self.AimPos, false).AbsRadAngle
			self.canFire[i] = true
			MovableMan:AddParticle(self.Barrel[i])
		else
			self.canFire[i] = false
			self.BarrelName[i] = ""
		end
	end
	
	if self:IsPlayerControlled() then
		self.player = true
		self.Target = nil
		
		-- Update ammo icon
		if not MovableMan:ValidMO(self.AmmoIcon) or self.AmmoIcon.PresetName ~= "Ammo Counter" then
			self.AmmoIcon = CreateMOSParticle("Ammo Counter", "EDV.rte")
			MovableMan:AddParticle(self.AmmoIcon)
		end
		
		self.AmmoIcon.Age = 1
		self.AmmoIcon.Pos = self.AboveHUDPos + Vector(0, -8)
		self.AmmoIcon.Frame = math.floor((self.ammo / self.maxAmmo) * self.AmmoIcon.FrameCount)
		
		-- Accelerate along the X-axis
		if self.Ctrl:IsState(Controller.MOVE_LEFT) and self.Vel.X > -10 then
			local Acc = Vector(-6000, 0)
			self:AddAbsForce(Acc, self.Pos)
		elseif self.Ctrl:IsState(Controller.MOVE_RIGHT) and self.Vel.X < 10 then
			local Acc = Vector(6000, 0)
			self:AddAbsForce(Acc, self.Pos)
		else
			self.AngularVel = self.AngularVel * 0.98
		end
		
		if self.Ctrl:IsMouseControlled() then			
			gunAim = self.Ctrl.AnalogAim.AbsRadAngle
		else	-- Keyboard
			if self.Ctrl:IsState(Controller.BODY_JUMP) then
				if self.HFlipped then
					self.aim = self.aim - 0.06
				else
					self.aim = self.aim + 0.06
				end
			elseif self.Ctrl:IsState(Controller.BODY_CROUCH) then
				if self.HFlipped then
					self.aim = self.aim + 0.06
				else
					self.aim = self.aim - 0.06
				end
			end
			
			if self.aim > 4.712389 then
				self.aim = self.aim - 6.283185
			elseif self.aim < -1.570796 then
				self.aim = self.aim + 6.283185
			end
			
			gunAim = self.RotAngle + self.aim
		end
		
		-- Fake sharp aim
		self.ViewPoint = Offset(self.Pos, gunAim, (FrameMan.PlayerScreenWidth + FrameMan.PlayerScreenHeight) / 6)
		
		-- Update the aim
		for i = 1, 7 do
			if not MovableMan:ValidMO(self.AimDots[i]) or self.AimDots[i].PresetName ~= "Aim" then
				self.AimDots[i] = CreateMOSParticle("Aim", "EDV.rte")
				MovableMan:AddParticle(self.AimDots[i])
			end
			
			self.AimDots[i].Vel = self.Vel
			self.AimDots[i].Pos = Offset(self.Pos, gunAim, 100 + i*25)
			self.AimDots[i].Age = 1
			
			if SceneMan.SceneWrapsX and self.AimDots[i].Pos.X > SceneMan.SceneWidth then
				self.AimDots[i].Pos.X = self.AimDots[i].Pos.X - SceneMan.SceneWidth
			end
		end
		
		self.traceAim = not self.traceAim
		if self.traceAim then
			SceneMan:CastObstacleRay(self.AimDots[1].Pos, Offset(nil, gunAim, 600), Vector(), self.AimPos, self.ID, self.Team, -1, 4) -- Check every fifth pixel
		end
		
		if self.canFire[3] and self.grenades < 1 and self.GrenadeTimer:IsPastSimMS(1900) and
			((self.Ctrl:IsMouseControlled() and LauncherKey()) or
			(not self.Ctrl:IsMouseControlled() and self.Ctrl:IsState(Controller.AIM_SHARP)))
		then
			self.grenades = 3
		elseif self.canFire[3] and self.grenades > 0 and self.GrenadeTimer:IsPastSimMS(120) then
			self.GrenadeTimer:Reset()
			
			local Grenade = CreateMOSRotating("Launcher Grenade", "EDV.rte")	-- A grenade spawned from an emitter can not inherit velocity
			local Flash = CreateAEmitter("Launcher Flash", "EDV.rte")
			local Muzzle
			
			if self.grenades == 3 then
				Muzzle = Offset(self.Offset[3], self.Barrel[3].RotAngle - 1.570796, 4)
			elseif self.grenades == 1 then
				Muzzle = Offset(self.Offset[3], self.Barrel[3].RotAngle + 1.570796, 4)
			else
				Muzzle = self.Offset[3]
			end
			
			Grenade.Pos = Offset(Muzzle, self.Barrel[3].RotAngle, 18)
			Grenade.Vel = Offset(Vector(0,0), self.Barrel[3].RotAngle + RangeRand(-0.05, 0.05), 28) + self.Vel
			Grenade.RotAngle = self.Barrel[3].RotAngle
			Grenade:SetWhichMOToNotHit(self.Barrel[3], -1)
			MovableMan:AddParticle(Grenade)
			
			Flash.Pos = Offset(Muzzle, self.Barrel[3].RotAngle, 12)
			Flash.Vel = self.Vel
			Flash.RotAngle = self.Barrel[3].RotAngle
			MovableMan:AddParticle(Flash)
			
			self.grenades = self.grenades - 1
		end
	elseif self.player then
		self.player = false	-- only execute this code once after the player has switched actor
		
		for i = 1, 7 do
			if MovableMan:ValidMO(self.AimDots[i]) and self.AimDots[i].PresetName == "Aim" then
				self.AimDots[i].ToDelete = true
			end
		end
		
		if MovableMan:ValidMO(self.AmmoIcon) then
			self.AmmoIcon.ToDelete = true
		end
	end
	
	-- Rotate left barrel
	local i = 1
	if self.canFire[i] then
		self.canFire[i] = false
		
		local brlAng = LimitAngle(self.Barrel[i].RotAngle)
		local aimAng = LimitAngle(SceneMan:ShortestDistance(self.Barrel[i].Pos, self.AimPos, false).AbsRadAngle)
		local brlQuad = math.floor(brlAng/1.571)	-- Calculate the quadrants
		local aimQuad = math.floor(aimAng/1.571)
		
		if brlQuad > aimQuad then
			brlAng = brlAng - 0.12
		elseif brlQuad < aimQuad then
			brlAng = brlAng + 0.12
		else
			if math.abs(brlAng - aimAng) < 0.07 then
				self.canFire[i] = true
			end
			
			if brlAng - aimAng > 0.03 then
				brlAng = brlAng - 0.06
			elseif brlAng - aimAng < -0.03 then
				brlAng = brlAng + 0.06
			end
		end
		
		brlAng = LimitAngle(brlAng)
		self.Barrel[i].RotAngle = math.min(math.max(brlAng, 1.3 + self.RotAngle), 5.5 + self.RotAngle)
	end
	
	-- Rotate right barrel
	local i = 2
	if self.canFire[i] then
		self.canFire[i] = false
		
		local brlAng = NormalizeAngle(self.Barrel[i].RotAngle)
		local aimAng = NormalizeAngle(SceneMan:ShortestDistance(self.Barrel[i].Pos, self.AimPos, false).AbsRadAngle)
		local brlQuad = math.floor((brlAng+3.142)/1.571)	-- Add pi to mirror the quadrant
		local aimQuad = math.floor((aimAng+3.142)/1.571)
		
		if brlQuad > aimQuad then
			brlAng = brlAng - 0.12
		elseif brlQuad < aimQuad then
			brlAng = brlAng + 0.12
		else
			if math.abs(brlAng - aimAng) < 0.07 then
				self.canFire[i] = true
			end
			
			if brlAng - aimAng > 0.03 then
				brlAng = brlAng - 0.06
			elseif brlAng - aimAng < -0.03 then
				brlAng = brlAng + 0.06
			end
		end
		
		brlAng = NormalizeAngle(brlAng)
		self.Barrel[i].RotAngle = math.max(math.min(brlAng, 1.84 + self.RotAngle), -2.35 + self.RotAngle)
	end
	
	-- Rotate the GL
	i = 3
	if self.canFire[i] then
		if gunAim > 1.571 then	-- Left side
			self.Barrel[i].RotAngle = math.max(math.min(gunAim, 3.07 + self.RotAngle), 1.87 + self.RotAngle)
		else	-- Right side
			self.Barrel[i].RotAngle = math.max(gunAim, 0.07 + self.RotAngle)
		end
	end
	
	if (self.canFire[1] or self.canFire[2]) and self.ammo > 0 and self.GunTimer:IsPastSimMS(self.laserRof) then
		if self.Ctrl:IsState(Controller.WEAPON_FIRE) or self.Target then
			self.GunTimer:Reset()
			self.RechargeTimer:Reset()
			self.laserRoT = self.laserStartRoT
			
			if self.laserRof == self.laserNormal then
				self.laserRof = self.laserBurst
				if self.canFire[1] then
					local Emitter = CreateAEmitter("Gunship Laser Emitter", "EDV.rte")
					Emitter.Pos = self.Barrel[1].Pos + self.Barrel[1]:RotateOffset(Vector(24, 4))
					Emitter.RotAngle = self.Barrel[1].RotAngle
					Emitter.Vel = self.Vel
					MovableMan:AddParticle(Emitter)
					self.ammo = self.ammo - 1
				end
				
				if self.canFire[2] and self.ammo > 0 then
					local Emitter = CreateAEmitter("Gunship Laser Emitter", "EDV.rte")
					Emitter.Pos = self.Barrel[2].Pos + self.Barrel[2]:RotateOffset(Vector(24, -4))
					Emitter.RotAngle = self.Barrel[2].RotAngle
					Emitter.Vel = self.Vel
					MovableMan:AddParticle(Emitter)
					self.ammo = self.ammo - 1
				end
			else
				self.laserRof = self.laserNormal
				if self.canFire[1] then
					local Emitter = CreateAEmitter("Gunship Laser Emitter", "EDV.rte")
					Emitter.Pos = self.Barrel[1].Pos + self.Barrel[1]:RotateOffset(Vector(24, -4))
					Emitter.RotAngle = self.Barrel[1].RotAngle
					Emitter.Vel = self.Vel
					MovableMan:AddParticle(Emitter)
					self.ammo = self.ammo - 1
				end
				
				if self.canFire[2] and self.ammo > 0 then
					local Emitter = CreateAEmitter("Gunship Laser Emitter", "EDV.rte")
					Emitter.Pos = self.Barrel[2].Pos + self.Barrel[2]:RotateOffset(Vector(24, 4))
					Emitter.RotAngle = self.Barrel[2].RotAngle
					Emitter.Vel = self.Vel
					MovableMan:AddParticle(Emitter)
					self.ammo = self.ammo - 1
				end
			end
		else
			self.laserRof = self.laserNormal
		end
	end
end

function UpdateAI(self)
	if self.PlayerInterferedTimer:IsPastSimTimeLimit() then
		local FuturePos = self.Pos + self.Vel*20
		
		-- Make sure FuturePos is inside the scene
		if FuturePos.X > SceneMan.SceneWidth then
			if SceneMan.SceneWrapsX then
				FuturePos.X = FuturePos.X - SceneMan.SceneWidth
			else
				FuturePos.X = SceneMan.SceneWidth - self.Radius
			end
		elseif FuturePos.X < 0 then
			if SceneMan.SceneWrapsX then
				FuturePos.X = FuturePos.X + SceneMan.SceneWidth
			else
				FuturePos.X = self.Radius
			end
		end
		
		local Dist = SceneMan:ShortestDistance(FuturePos, self.LZpos, false)
		if math.abs(Dist.X) > 100 then
			if self.DeliveryState == ACraft.LAUNCH then
				self.LZpos.X = FuturePos.X
			else
				self.LZpos = SceneMan:MovePointToGround(FuturePos, self.Radius*2.5, 5)
			end
		end
	end
	
	self.PlayerInterferedTimer:Reset()
	
	if self.AIMode ~= self.LastAIMode then
		self.LastAIMode = self.AIMode

		if self.AIMode == Actor.AIMODE_RETURN then
			self.DeliveryState = ACraft.LAUNCH
			self.LZpos = Vector(self.Pos.X, -500)	-- Go to orbit
		else -- Actor.AIMODE_STAY and Actor.AIMODE_DELIVER
			local FuturePos = self.Pos + self.Vel*20
			
			-- Make sure FuturePos is inside the scene
			if FuturePos.X > SceneMan.SceneWidth then
				if SceneMan.SceneWrapsX then
					FuturePos.X = FuturePos.X - SceneMan.SceneWidth
				else
					FuturePos.X = SceneMan.SceneWidth - self.Radius
				end
			elseif FuturePos.X < 0 then
				if SceneMan.SceneWrapsX then
					FuturePos.X = FuturePos.X + SceneMan.SceneWidth
				else
					FuturePos.X = self.Radius
				end
			end
			
			self.LZpos = SceneMan:MovePointToGround(FuturePos, self.Radius*2.5, 5)
			self.DeliveryState = ACraft.FALL
		end
	end
	
	-- Control right/left movement
	local dist = SceneMan:ShortestDistance(self.Pos+self.Vel*15, self.LZpos, false).X
	local change = self.XposPID:Update(dist, 0)
	if change > 3 and self.Vel.X < 15 then
		self.Ctrl.AnalogMove = Vector(change/30, 0)
	elseif change < -3 and self.Vel.X > -15 then
		self.Ctrl.AnalogMove = Vector(change/30, 0)
	end

	-- Control up/down movement
	change = self.YposPD:Update(self.Vel.Y-(self.LZpos.Y-self.Pos.Y), 0)
	if change > 1 then
		self.AltitudeMoveState = ACraft.ASCEND
	elseif change < -1 then
		self.AltitudeMoveState = ACraft.DESCEND
	end

	-- Delivery Sequence logic
	if self.DeliveryState == ACraft.FALL then
		-- Don't descend if we have nothing to deliver
		if self:IsInventoryEmpty() then
			if self.AIMode ~= Actor.AIMODE_STAY then
				self.DeliveryState = ACraft.LAUNCH
				self.HatchTimer:Reset()
				self.LZpos.Y = -500	-- Go to orbit
			end
		else
			dist = SceneMan:ShortestDistance(self.Pos, self.LZpos, false).Magnitude
			if dist < 20 and math.abs(change) < 2 and self.Vel.Magnitude < 4 then	-- If we passed the hover check, start unloading
				if self.AIMode ~= Actor.AIMODE_STAY then
					self.AltitudeMoveState = ACraft.HOVER
				end
			else	-- Check for something in the way of our descent, and hover to the side to avoid it
				self.search = not self.search	-- Search every second update
				if self.search then
					local obstID = self:DetectObstacle(self.Radius * 7)
					if obstID > 0 then
						local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(obstID))
						if MO.ClassName == "ACDropShip" or MO.ClassName == "ACRocket" then
							self.LZpos.X = self.LZpos.X + 15
							self.AltitudeMoveState = ACraft.HOVER
						end
					end
				else
					local Free = Vector()
					local Start = self.Pos + Vector(self.Radius, 0)
					local Trace = self.Vel * (self.Radius/4) + Vector(0,50)
					if PosRand() < 0.5 then
						Start.X = Start.X - self.Diameter
					end
					
					if SceneMan:CastStrengthRay(Start, Trace, 0, Free, 4, 0, true) then
						self.LZpos.X = self.Pos.X
						self.LZpos.Y = Free.Y - self.Radius*2.5
					end
				end
			end
		end
	elseif self.DeliveryState == ACraft.LAUNCH then
		-- Check for something in the way of our ascent, and hover to the side to avoid it
		self.search = not self.search	-- Search every second update
		if self.search then
			local obstID = self:DetectObstacle(self.Radius * 7)
			if obstID > 0 then
				local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(obstID))
				if MO.ClassName == "ACDropShip" or MO.ClassName == "ACRocket" then
					self.LZpos.X = self.LZpos.X - 15
					self.AltitudeMoveState = ACraft.HOVER
				end
			end
		end
	end

	-- If we are hopelessly stuck, self destruct
	if self.Vel.Magnitude > 1 then
		self.StuckTimer:Reset()
	elseif self.StuckTimer:IsPastSimMS(30000) then
		self:GibThis()
	end

	-- Input translation
	if self.AltitudeMoveState == ACraft.ASCEND then
		self.Ctrl:SetState(Controller.MOVE_UP, true)
	elseif self.AltitudeMoveState == ACraft.DESCEND then
		self.Ctrl:SetState(Controller.MOVE_DOWN, true)
	else
		self.Ctrl:SetState(Controller.MOVE_UP, false)
		self.Ctrl:SetState(Controller.MOVE_DOWN, false)
	end
	
	if not self.findTarget then
		-- Look for targets
		if self.canFire[1] or self.canFire[2] then
			self.findTarget = 0
			self.Target = false
			local Dist, i
			local range = (FrameMan.PlayerScreenWidth + FrameMan.PlayerScreenHeight) / 2.5
			for Act in MovableMan.Actors do
				if Act.Team ~= self.Team then
					Dist = SceneMan:ShortestDistance(self.Pos, Act.Pos, false)
					if Dist.Magnitude < range then
						if Dist.AbsRadAngle < 1.571 and self.canFire[1] then
							i = 1
						elseif self.canFire[2] then
							i = 2
						elseif self.canFire[1] then
							i = 1
						end
						
						if not MovableMan:ValidMO(self.Barrel[i]) then
							self.canFire[i] = false
							self.Barrel[i] = nil
							break
						end
						
						Dist = SceneMan:ShortestDistance(self.Barrel[i].Pos, Act.Pos, false)
						if SceneMan:CastFindMORay(self.Barrel[i].Pos, Dist, Act.ID, self.AimPos, self.Barrel[i].ID, false, 4) then
							self.Target = true
							break
						end
					end
				end
			end
		end
	else
		self.findTarget = self.findTarget + 1
		if self.findTarget > 15 then
			self.findTarget = nil
			
			-- Look for waypoints
			Dist = SceneMan:ShortestDistance(self.Pos, self:GetLastAIWaypoint(), false)
			if Dist.Magnitude > 50 then
				self.LZpos = self.Pos+Dist
			end
		end
	end
end
