function Create(self)
	self.Ctrl = self:GetController()
	self.Target = Vector()
	self.Aim = {}
	self.aimAng = 0
	self.range = FrameMan.PlayerScreenWidth * 0.5
	
	for i = 1, 3 do
		self.Aim[i] = CreateMOPixel("Hover Tank Aim", "EDV.rte")
	end
	
	-- Find weapon
	local MoObj
	for i = 1, MovableMan:GetMOIDCount()-1 do
		if MovableMan:GetRootMOID(i) == self.ID then
			MoObj = MovableMan:GetMOFromID(i)
			if MoObj.ClassName == "HDFirearm" then
				self.Barrel = ToAttachable(MoObj)
			elseif MoObj.ClassName == "Turret" then
				self.Turret = ToAttachable(MoObj)
			end
		end
	end
	
	--[[ Global
	if not gHoverTank then
		gHoverTank = {}
	end]]
end

function Destroy(self)
	for i = 1, #self.Aim do
		if MovableMan:ValidMO(self.Aim[i].Pt) then
			self.Aim[i].Pt.ToDelete = true
		end
	end
	
	if MovableMan:ValidMO(self.Homing) then
		self.Homing.ToDelete = true
	end
end

function Update(self)
	if self.Health < 1 then
		self:GibThis()
	end
	
	if not (not self:IsPlayerControlled() and self.AIMode == Actor.AIMODE_SENTRY) then	-- Don't move when the AI is in sentry mode.
		local alt = 120	-- The engines won't generate any lift above this altitude.
		local front = math.min(SceneMan:FindAltitude(self.Pos + self:RotateOffset(Vector(35, 20)), alt, 5), alt)
		local rear = math.min(SceneMan:FindAltitude(self.Pos + self:RotateOffset(Vector(-35, 20)), alt, 5), alt)
		
		-- Movement
		if self.Ctrl:IsState(Controller.MOVE_RIGHT) or self.Ctrl:IsState(Controller.MOVE_LEFT) then
			if (front + rear) < alt*1.9 then
				self:AddForce(self:RotateOffset(Vector(8000, 0)), Vector(0, 0))
			else
				self:AddForce(self:RotateOffset(Vector(3000, 0)), Vector(0, 0))
			end
		else
			self.Vel.X = self.Vel.X * 0.98
		end
		
		local power = 6000
		if self.Vel.Y < -1 then
			power = power * 0.92
		elseif self.Vel.Y > 7 then
			power = power * 1.1
		end
		
		if self.Ctrl:IsState(Controller.BODY_JUMP) then
			power = power * 1.55
			local Smoke = CreateMOSRotating("Engine Wash", "EDV.rte")
			if Smoke then
				Smoke.Pos = self.Pos + self:RotateOffset(Vector(SelectRand(-35,35), 22))
				Smoke.Vel = self:RotateOffset(Vector(RangeRand(-2,2), 12))
				MovableMan:AddParticle(Smoke)
			end
		elseif self.Ctrl:IsState(Controller.BODY_CROUCH) then
			power = power * 0.25
		end
		
		if front > rear then
			if self.HFlipped then
				if self.Vel.X < -5 then
					rear = (front + rear)/2
				end
			elseif self.Vel.X > 5 then
				rear = (front + rear)/2
			end
		end
		
		-- Apply a larger force for an engine close to the ground. This will keep the tank uppright.
		self:AddForce(self:RotateOffset(Vector(-100, -power*((1.25-rear/alt)^2.3))), self:RotateOffset(Vector(-1.7, 0.7)))
		self:AddForce(self:RotateOffset(Vector(100, -power*((1.25-front/alt)^2.3))), self:RotateOffset(Vector(1.7, 0.7)))
		
		-- Reduce the risk of flipping.
		if (front > alt-1 and rear > alt-1) or math.abs(self.RotAngle) > 1 then
			self.RotAngle = self.RotAngle * 0.91
		else
			self.RotAngle = self.RotAngle * 0.96
		end
	end
	
	if not self.Homing then
		self.Barrel.Sharpness = 1	-- Don't home in on the target marker any more.
	end
	
	if self.Turret:IsAttached() then
		if self.Barrel:IsAttached() then	-- Aim
			SceneMan:CastObstacleRay(self.Barrel.Pos + self.Barrel:RotateOffset(Vector(15,0)), self.Barrel:RotateOffset(Vector(self.range,0)), Vector(), self.Target, self.ID, self.IgnoresWhichTeam, -1, 4) -- Check every fifth pixel
			self.Target = self.Target + self.Barrel:RotateOffset(Vector(-8,0))	-- Draw the aim in front of the target
			self.aimAng = self.aimAng + 0.2
			for i = 1, #self.Aim do
				if not MovableMan:ValidMO(self.Aim[i].Pt) or self.Aim[i].Pt.PresetName ~= "Hover Tank Aim" then
					self.Aim[i].Pt = CreateMOPixel("Hover Tank Aim", "EDV.rte")
					MovableMan:AddParticle(self.Aim[i].Pt)
				end
				
				local ang = self.aimAng + i*6.28/3
				self.Aim[i].Pt.Pos = self.Target + Vector(3*math.cos(ang) - 3*math.sin(ang), 3*math.sin(ang) + 3*math.cos(ang))
				self.Aim[i].Pt.Age = 1
			end
			
			-- Fire
			if self.Ctrl:IsState(Controller.WEAPON_FIRE) then
				self.Barrel.Sharpness = SceneMan:ShortestDistance(self.Barrel.Pos, self.Target, false).Magnitude
				if not MovableMan:ValidMO(self.Homing) then
					self.Homing = CreateMOPixel("Hover Tank Homing", "EDV.rte")
					MovableMan:AddParticle(self.Homing)
				end
				
				self.Homing.Pos = self.Target
			elseif self.Barrel.Sharpness > 1 and MovableMan:ValidMO(self.Homing) then
				self.Homing.ToDelete = true
				self.Homing = nil
			end
		else
			for i = 1, #self.Aim do
				if MovableMan:ValidMO(self.Aim[i].Pt) then
					self.Aim[i].Pt.ToDelete = true
				end
			end
		end
	else
		self:GibThis()
	end
end
