function Create(self)
	self.Life = Timer()
	self.timing = 200	-- arm time
	self.proximity = nil
	self.armDist = nil
	self.Target = nil
	
	self.Life:Reset()
end

function Update(self)
	local speed = self.Vel.Magnitude * math.cos(self.RotAngle - self.Vel.AbsRadAngle)    -- Movement in direction of view angle
	if self.Vel.Magnitude > speed * 1.1 and self.Vel.Magnitude > 10 then	-- Fake extra resistance when moving perpendicular to the view angle
		local Tmp = Vector(self.Vel.X, self.Vel.Y)
		Tmp:RadRotate(-self.Vel.AbsRadAngle)
		Tmp:SetXY(Tmp.X*0.97, Tmp.Y)
		Tmp:RadRotate(self.Vel.AbsRadAngle)
		self.Vel = Tmp
	end
	
	self.AngularVel = self.AngularVel * 0.97
	
	if self.Life:IsPastSimMS(self.timing) then	-- arm the missile		
		self.Life:Reset()
		
		if self.Target == nil then
			if self.Sharpness < 255 then	-- Get target ID
				self.Target = MovableMan:GetMOFromID(self.Sharpness)
				self.Sharpness = 255
				if MovableMan:ValidMO(self.Target) then
					self.armDist = self.Target.Radius + 35
				else
					self.Target = nil
				end
			elseif speed < 8 and self.timing > 200 then
				self:GibThis()
			end
		end
		
		self.HitsMOs = true
		self.timing = 500
	end
	
	if self.Target ~= nil then
		if MovableMan:ValidMO(self.Target) then
			local range = (self.Pos - self.Target.Pos).Magnitude	-- X Wrap?
			if self.proximity ~= nil and range > self.proximity then
				self:GibThis()
				ToActor(self.Target):GibThis() -- cheat; gib the target
				self.Target = nil
			else
				if range < self.armDist then
					self.proximity = range		-- the missile is close, arm the proximity fuse
				end
				
				local AimPoint = nil
				local t2t = math.min((range / speed) * 0.8, 30)	-- estimated time to target
				if t2t > 7 then
					AimPoint = self.Target.Pos + self.Target.Vel * t2t
					AimPoint.Y = math.max(math.min(AimPoint.Y, SceneMan.SceneHeight), 0)
				else
					AimPoint = self.Target.Pos
				end
				
				local diff = (AimPoint - (self.Pos + self.Vel * (t2t/2))).AbsRadAngle - self.RotAngle
				if diff > 3.141592 then
					diff = diff - math.pi*2
				elseif diff < -3.141592 then
					diff = diff + math.pi*2
				end
				
				local absDiff = math.abs(diff)
				if absDiff < 2 then	-- inside a ~200 degree cone
					self.GetThrottle = 0.74*absDiff^2 - 2.33*absDiff + 1	-- set throttle based on angle to target
					if absDiff < 0.09 then	-- within a 10 degree cone
						self.EmitAngle = 3.141592
						if speed < 20 and range > 300 then
							self.GetThrottle = 1.0
						end
					else
						self.EmitAngle = 3.141592 + diff * 0.8	-- vector thrusters
					end
					
					if diff > 0 then	 -- slowly rotate towards the target
						self.RotAngle = self.RotAngle + math.min(diff, 0.06)
					elseif diff < 0 then
						self.RotAngle = self.RotAngle + math.max(diff, -0.06)
					end
				else
					self.GetThrottle = -0.5
					self.EmitAngle = 3.141592
				end
				
				if speed > 65 then	-- limit speed
					self.GetThrottle = math.min(self.GetThrottle, -0.5)
				end
			end
		else
			self:GibThis()
			self.Target = nil
		end
	end
end
